@@ -26,24 +26,79 @@
 
 #include <tesseract_common/macros.h>
 TESSERACT_COMMON_IGNORE_WARNINGS_PUSH
+#include <boost/graph/adj_list_serialize.hpp>
 #include <boost/graph/dijkstra_shortest_paths.hpp>
 #include <fstream>
 #include <queue>
 #include <console_bridge/console.h>
+#include <boost/graph/undirected_graph.hpp>
+#include <boost/graph/copy.hpp>
+#include <boost/serialization/access.hpp>
+#include <boost/serialization/base_object.hpp>
+#include <boost/serialization/nvp.hpp>
+#include <boost/serialization/shared_ptr.hpp>
+#include <boost/serialization/split_member.hpp>
+#if (BOOST_VERSION >= 107400) && (BOOST_VERSION < 107500)
+#include <boost/serialization/library_version_type.hpp>
+#endif
+#include <boost/serialization/unordered_map.hpp>
+#include <boost/serialization/utility.hpp>
 TESSERACT_COMMON_IGNORE_WARNINGS_POP
 
 #include <tesseract_scene_graph/graph.h>
+#include <tesseract_common/utils.h>
 
 namespace tesseract_scene_graph
 {
-SceneGraph::SceneGraph(const std::string& name) : acm_(std::make_shared<AllowedCollisionMatrix>())
+using UGraph =
+    boost::adjacency_list<boost::listS, boost::listS, boost::undirectedS, VertexProperty, EdgeProperty, GraphProperty>;
+
+struct ugraph_vertex_copier
+{
+  ugraph_vertex_copier(const Graph& g1, UGraph& g2)
+    : vertex_all_map1(get(boost::vertex_all, g1)), vertex_all_map2(get(boost::vertex_all, g2))
+  {
+  }
+
+  template <typename Vertex1, typename Vertex2>
+  void operator()(const Vertex1& v1, Vertex2& v2) const
+  {
+    boost::put(vertex_all_map2, v2, get(vertex_all_map1, v1));
+  }
+  typename boost::property_map<Graph, boost::vertex_all_t>::const_type vertex_all_map1;
+  mutable typename boost::property_map<UGraph, boost::vertex_all_t>::type vertex_all_map2;
+};
+
+SceneGraph::SceneGraph(const std::string& name) : acm_(std::make_shared<tesseract_common::AllowedCollisionMatrix>())
 {
   boost::set_property(static_cast<Graph&>(*this), boost::graph_name, name);
 }
 
-SceneGraph::Ptr SceneGraph::clone() const
+SceneGraph::SceneGraph(SceneGraph&& other) noexcept
+  : Graph(std::forward<Graph>(other))
+  , link_map_(std::move(other.link_map_))
+  , joint_map_(std::move(other.joint_map_))
+  , acm_(std::move(other.acm_))
 {
-  SceneGraph::Ptr cloned_graph = std::make_shared<SceneGraph>();
+  rebuildLinkAndJointMaps();
+}
+
+SceneGraph& SceneGraph::operator=(SceneGraph&& other) noexcept
+{
+  Graph::operator=(std::forward<Graph>(other));
+
+  link_map_ = std::move(other.link_map_);
+  joint_map_ = std::move(other.joint_map_);
+  acm_ = std::move(other.acm_);
+
+  rebuildLinkAndJointMaps();
+
+  return *this;
+}
+
+SceneGraph::UPtr SceneGraph::clone() const
+{
+  auto cloned_graph = std::make_unique<SceneGraph>();
 
   for (auto& link : getLinks())
   {
@@ -118,19 +173,16 @@ bool SceneGraph::addLink(const Link& link, const Joint& joint)
     return false;
   }
 
-  std::string link_name = link.getName();
-  std::string joint_name = joint.getName();
-
   if (!addLinkHelper(std::make_shared<Link>(link.clone())))
-    return false;
+    return false;  // LCOV_EXCL_LINE
 
   if (!addJointHelper(std::make_shared<Joint>(joint.clone())))
-    return false;
+    return false;  // LCOV_EXCL_LINE
 
   return true;
 }
 
-bool SceneGraph::addLinkHelper(Link::Ptr link_ptr, bool replace_allowed)
+bool SceneGraph::addLinkHelper(const Link::Ptr& link_ptr, bool replace_allowed)
 {
   auto found = link_map_.find(link_ptr->getName());
   bool link_exists = (found != link_map_.end());
@@ -179,7 +231,20 @@ std::vector<Link::ConstPtr> SceneGraph::getLinks() const
   return links;
 }
 
-bool SceneGraph::removeLink(const std::string& name)
+std::vector<Link::ConstPtr> SceneGraph::getLeafLinks() const
+{
+  std::vector<Link::ConstPtr> links;
+  links.reserve(link_map_.size());
+  for (const auto& link : link_map_)
+  {
+    if (boost::out_degree(link.second.second, *this) == 0)
+      links.push_back(link.second.first);
+  }
+
+  return links;
+}
+
+bool SceneGraph::removeLink(const std::string& name, bool recursive)
 {
   auto found = link_map_.find(name);
   if (found == link_map_.end())
@@ -188,7 +253,9 @@ bool SceneGraph::removeLink(const std::string& name)
     return false;
   }
 
-  // Needt to remove all inbound and outbound edges first
+  std::vector<std::string> adjacent_link_names = getAdjacentLinkNames(name);
+
+  // Need to remove all inbound and outbound edges first
   Vertex vertex = getVertex(name);
   boost::clear_vertex(vertex, *this);
 
@@ -209,9 +276,42 @@ bool SceneGraph::removeLink(const std::string& name)
   // Need to remove any reference to link in allowed collision matrix
   removeAllowedCollision(name);
 
+  if (recursive)
+  {
+    for (const auto& link_name : adjacent_link_names)
+    {
+      if (getInboundJoints(link_name).empty())
+        removeLink(link_name, true);
+    }
+  }
+
   return true;
 }
 
+bool SceneGraph::moveLink(const Joint& joint)
+{
+  if (link_map_.find(joint.child_link_name) == link_map_.end())
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to move link (%s) in scene graph that does not exist.",
+                           joint.child_link_name.c_str());
+    return false;
+  }
+
+  if (link_map_.find(joint.parent_link_name) == link_map_.end())
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to move link (%s) in scene graph that parent link (%s) which does not exist.",
+                           joint.child_link_name.c_str(),
+                           joint.parent_link_name.c_str());
+    return false;
+  }
+
+  std::vector<tesseract_scene_graph::Joint::ConstPtr> joints = getInboundJoints(joint.child_link_name);
+  for (const auto& joint : joints)
+    removeJoint(joint->getName());
+
+  return addJoint(joint);
+}
+
 void SceneGraph::setLinkVisibility(const std::string& name, bool visibility)
 {
   boost::property_map<Graph, boost::vertex_link_visible_t>::type param =
@@ -246,7 +346,7 @@ bool SceneGraph::addJoint(const Joint& joint)
   return addJointHelper(joint_ptr);
 }
 
-bool SceneGraph::addJointHelper(Joint::Ptr joint_ptr)
+bool SceneGraph::addJointHelper(const Joint::Ptr& joint_ptr)
 {
   auto parent = link_map_.find(joint_ptr->parent_link_name);
   auto child = link_map_.find(joint_ptr->child_link_name);
@@ -270,6 +370,30 @@ bool SceneGraph::addJointHelper(Joint::Ptr joint_ptr)
     return false;
   }
 
+  if ((joint_ptr->type != JointType::FIXED) && (joint_ptr->type != JointType::FLOATING) &&
+      (joint_ptr->type != JointType::CONTINUOUS) && joint_ptr->limits == nullptr)
+  {
+    CONSOLE_BRIDGE_logWarn("Joint with name (%s) requires limits because it is not of type fixed, floating or "
+                           "continuous.",
+                           joint_ptr->getName().c_str());
+    return false;
+  }
+
+  // Need to set limits for continuous joints. TODO: This may not be required
+  // by the optimization library but may be nice to have
+  if (joint_ptr->type == tesseract_scene_graph::JointType::CONTINUOUS)
+  {
+    if (joint_ptr->limits == nullptr)
+    {
+      joint_ptr->limits = std::make_shared<JointLimits>(-4 * M_PI, 4 * M_PI, 0, 2, 1);
+    }
+    else if (tesseract_common::almostEqualRelativeAndAbs(joint_ptr->limits->lower, joint_ptr->limits->upper, 1e-5))
+    {
+      joint_ptr->limits->lower = -4 * M_PI;
+      joint_ptr->limits->upper = +4 * M_PI;
+    }
+  }
+
   double d = joint_ptr->parent_to_joint_origin_transform.translation().norm();
 
   EdgeProperty info(joint_ptr, d);
@@ -290,29 +414,50 @@ Joint::ConstPtr SceneGraph::getJoint(const std::string& name) const
   return found->second.first;
 }
 
-bool SceneGraph::removeJoint(const std::string& name)
+bool SceneGraph::removeJoint(const std::string& name, bool recursive)
 {
   auto found = joint_map_.find(name);
   if (found == joint_map_.end())
     return false;
 
-  boost::remove_edge(found->second.second, static_cast<Graph&>(*this));
-  joint_map_.erase(name);
+  if (!recursive)
+  {
+    boost::remove_edge(found->second.second, static_cast<Graph&>(*this));
+    joint_map_.erase(name);
+  }
+  else
+  {
+    if (getInboundJoints(found->second.first->child_link_name).size() == 1)
+    {
+      std::string child_link_name = found->second.first->child_link_name;
+      removeLink(child_link_name, true);
+    }
+  }
 
   return true;
 }
 
 bool SceneGraph::moveJoint(const std::string& name, const std::string& parent_link)
 {
-  auto found = joint_map_.find(name);
+  auto found_joint = joint_map_.find(name);
+  auto found_parent_link = link_map_.find(parent_link);
 
-  if (found == joint_map_.end())
+  if (found_joint == joint_map_.end())
   {
     CONSOLE_BRIDGE_logWarn("Tried to move Joint with name (%s) which does not exist in scene graph.", name.c_str());
     return false;
   }
 
-  Joint::Ptr joint = found->second.first;
+  if (found_parent_link == link_map_.end())
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to move Joint with name (%s) to parent link (%s) which does not exist in scene "
+                           "graph.",
+                           name.c_str(),
+                           parent_link.c_str());
+    return false;
+  }
+
+  Joint::Ptr joint = found_joint->second.first;
   if (!removeJoint(name))
     return false;
 
@@ -330,6 +475,19 @@ std::vector<Joint::ConstPtr> SceneGraph::getJoints() const
   return joints;
 }
 
+std::vector<Joint::ConstPtr> SceneGraph::getActiveJoints() const
+{
+  std::vector<Joint::ConstPtr> joints;
+  joints.reserve(joint_map_.size());
+  for (const auto& joint : joint_map_)
+  {
+    if ((joint.second.first->type != JointType::FIXED) && (joint.second.first->type != JointType::FLOATING))
+      joints.push_back(joint.second.first);
+  }
+
+  return joints;
+}
+
 bool SceneGraph::changeJointOrigin(const std::string& name, const Eigen::Isometry3d& new_origin)
 {
   auto found = joint_map_.find(name);
@@ -364,9 +522,9 @@ bool SceneGraph::changeJointLimits(const std::string& name, const JointLimits& l
     return false;
   }
 
-  if (found->second.first->type == JointType::FIXED)
+  if (found->second.first->type == JointType::FIXED || found->second.first->type == JointType::FLOATING)
   {
-    CONSOLE_BRIDGE_logWarn("Tried to change Joint limits for a fixed joint type.", name.c_str());
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint limits for a fixed or floating joint type.", name.c_str());
     return false;
   }
 
@@ -382,6 +540,77 @@ bool SceneGraph::changeJointLimits(const std::string& name, const JointLimits& l
   return true;
 }
 
+bool SceneGraph::changeJointPositionLimits(const std::string& name, double lower, double upper)
+{
+  auto found = joint_map_.find(name);
+
+  if (found == joint_map_.end())
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint Position limits with name (%s) which does not exist in scene graph.",
+                           name.c_str());
+    return false;
+  }
+
+  if (found->second.first->type == JointType::FIXED || found->second.first->type == JointType::FLOATING)
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint Position limits for a fixed or floating joint type.", name.c_str());
+    return false;
+  }
+
+  found->second.first->limits->lower = lower;
+  found->second.first->limits->upper = upper;
+
+  return true;
+}
+
+bool SceneGraph::changeJointVelocityLimits(const std::string& name, double limit)
+{
+  auto found = joint_map_.find(name);
+
+  if (found == joint_map_.end())
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint Velocity limit with name (%s) which does not exist in scene graph.",
+                           name.c_str());
+    return false;
+  }
+
+  if (found->second.first->type == JointType::FIXED || found->second.first->type == JointType::FLOATING)
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint Velocity limit for a fixed or floating joint type.", name.c_str());
+    return false;
+  }
+
+  found->second.first->limits->velocity = limit;
+  return true;
+}
+
+bool SceneGraph::changeJointAccelerationLimits(const std::string& name, double limit)
+{
+  auto found = joint_map_.find(name);
+
+  if (found == joint_map_.end())
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint Acceleration limit with name (%s) which does not exist in scene "
+                           "graph.",
+                           name.c_str());
+    return false;
+  }
+
+  if (found->second.first->type == JointType::FIXED || found->second.first->type == JointType::FLOATING)
+  {
+    CONSOLE_BRIDGE_logWarn("Tried to change Joint Acceleration limit for a fixed or floating joint type.",
+                           name.c_str());
+    return false;
+  }
+
+  if (found->second.first->limits == nullptr)
+    found->second.first->limits = std::make_shared<JointLimits>();
+
+  found->second.first->limits->acceleration = limit;
+
+  return true;
+}
+
 JointLimits::ConstPtr SceneGraph::getJointLimits(const std::string& name)
 {
   auto found = joint_map_.find(name);
@@ -417,9 +646,9 @@ bool SceneGraph::isCollisionAllowed(const std::string& link_name1, const std::st
   return acm_->isCollisionAllowed(link_name1, link_name2);
 }
 
-AllowedCollisionMatrix::ConstPtr SceneGraph::getAllowedCollisionMatrix() const { return acm_; }
+tesseract_common::AllowedCollisionMatrix::ConstPtr SceneGraph::getAllowedCollisionMatrix() const { return acm_; }
 
-AllowedCollisionMatrix::Ptr SceneGraph::getAllowedCollisionMatrix() { return acm_; }
+tesseract_common::AllowedCollisionMatrix::Ptr SceneGraph::getAllowedCollisionMatrix() { return acm_; }
 
 Link::ConstPtr SceneGraph::getSourceLink(const std::string& joint_name) const
 {
@@ -440,7 +669,7 @@ std::vector<Joint::ConstPtr> SceneGraph::getInboundJoints(const std::string& lin
   std::vector<Joint::ConstPtr> joints;
   Vertex vertex = getVertex(link_name);
 
-  // Get incomming edges
+  // Get incoming edges
   auto num_in_edges = static_cast<int>(boost::in_degree(vertex, *this));
   if (num_in_edges == 0)  // The root of the tree will have not incoming edges
     return joints;
@@ -460,7 +689,7 @@ std::vector<Joint::ConstPtr> SceneGraph::getOutboundJoints(const std::string& li
   std::vector<Joint::ConstPtr> joints;
   Vertex vertex = getVertex(link_name);
 
-  // Get incomming edges
+  // Get incoming edges
   auto num_out_edges = static_cast<int>(boost::out_degree(vertex, *this));
   if (num_out_edges == 0)
     return joints;
@@ -517,7 +746,7 @@ std::vector<std::string> SceneGraph::getAdjacentLinkNames(const std::string& nam
 {
   std::vector<std::string> link_names;
   Vertex v = getVertex(name);
-  for (auto vd : boost::make_iterator_range(adjacent_vertices(v, *this)))
+  for (auto* vd : boost::make_iterator_range(adjacent_vertices(v, *this)))
     link_names.push_back(boost::get(boost::vertex_link, *this)[vd]->getName());
 
   return link_names;
@@ -527,7 +756,7 @@ std::vector<std::string> SceneGraph::getInvAdjacentLinkNames(const std::string&
 {
   std::vector<std::string> link_names;
   Vertex v = getVertex(name);
-  for (auto vd : boost::make_iterator_range(inv_adjacent_vertices(v, *this)))
+  for (auto* vd : boost::make_iterator_range(inv_adjacent_vertices(v, *this)))
     link_names.push_back(boost::get(boost::vertex_link, *this)[vd]->getName());
 
   return link_names;
@@ -551,6 +780,48 @@ std::vector<std::string> SceneGraph::getJointChildrenNames(const std::string& na
   return getLinkChildrenHelper(v);  // NOLINT
 }
 
+std::vector<std::string> SceneGraph::getJointChildrenNames(const std::vector<std::string>& names) const
+{
+  std::set<std::string> link_names;
+  for (const auto& name : names)
+  {
+    std::vector<std::string> joint_children = getJointChildrenNames(name);
+    link_names.insert(joint_children.begin(), joint_children.end());
+  }
+
+  return std::vector<std::string>{ link_names.begin(), link_names.end() };
+}
+
+std::unordered_map<std::string, std::string>
+SceneGraph::getAdjacencyMap(const std::vector<std::string>& link_names) const
+{
+  std::map<Vertex, size_t> index_map;
+  boost::associative_property_map<std::map<Vertex, size_t>> prop_index_map(index_map);
+
+  std::map<Vertex, boost::default_color_type> color_map;
+  boost::associative_property_map<std::map<Vertex, boost::default_color_type>> prop_color_map(color_map);
+
+  int c = 0;
+  Graph::vertex_iterator i, iend;
+  for (boost::tie(i, iend) = boost::vertices(*this); i != iend; ++i, ++c)
+    boost::put(prop_index_map, *i, c);
+
+  std::unordered_map<std::string, std::string> adjacency_map;
+  for (const auto& link_name : link_names)
+  {
+    Vertex start_vertex = getVertex(link_name);
+    adjacency_detector vis(adjacency_map, color_map, link_name, link_names);
+
+    // NOLINTNEXTLINE
+    boost::breadth_first_search(
+        *this,
+        start_vertex,
+        boost::visitor(vis).root_vertex(start_vertex).vertex_index_map(prop_index_map).color_map(prop_color_map));
+  }
+
+  return adjacency_map;
+}
+
 void SceneGraph::saveDOT(const std::string& path) const
 {
   std::ofstream dot_file(path);
@@ -582,78 +853,112 @@ void SceneGraph::saveDOT(const std::string& path) const
   dot_file << "}";
 }
 
-SceneGraph::Path SceneGraph::getShortestPath(const std::string& root, const std::string& tip) const
+ShortestPath SceneGraph::getShortestPath(const std::string& root, const std::string& tip) const
 {
-  const Graph& graph = *this;
-  Vertex s = getVertex(root);
+  // Must copy to undirected graph because order does not matter for creating kinematics chains.
 
-  std::map<Vertex, Vertex> predicessor_map;
-  boost::associative_property_map<std::map<Vertex, Vertex>> prop_predicessor_map(predicessor_map);
+  // Copy Graph
+  UGraph graph;
 
-  std::map<Vertex, double> distance_map;
-  boost::associative_property_map<std::map<Vertex, double>> prop_distance_map(distance_map);
+  std::map<Graph::vertex_descriptor, size_t> index_map;
+  boost::associative_property_map<std::map<Graph::vertex_descriptor, size_t>> prop_index_map(index_map);
 
-  std::map<Vertex, size_t> index_map;
-  boost::associative_property_map<std::map<Vertex, size_t>> prop_index_map(index_map);
+  {
+    int c = 0;
+    Graph::vertex_iterator i, iend;
+    for (boost::tie(i, iend) = boost::vertices(*this); i != iend; ++i, ++c)
+      boost::put(prop_index_map, *i, c);
+  }
 
-  int c = 0;
-  Graph::vertex_iterator i, iend;
-  for (boost::tie(i, iend) = boost::vertices(graph); i != iend; ++i, ++c)
-    boost::put(prop_index_map, *i, c);
+  ugraph_vertex_copier v_copier(*this, graph);
+  boost::copy_graph(*this, graph, boost::vertex_index_map(prop_index_map).vertex_copy(v_copier));
 
-  std::map<Edge, double> weight_map;
-  boost::associative_property_map<std::map<Edge, double>> prop_weight_map(weight_map);
-  Graph::edge_iterator j, jend;
-  for (boost::tie(j, jend) = boost::edges(graph); j != jend; ++j)
-    boost::put(prop_weight_map, *j, boost::get(boost::edge_weight, graph)[*j]);
+  // Search Graph
+  UGraph::vertex_descriptor s_root = getVertex(root);
+  UGraph::vertex_descriptor s_tip = getVertex(tip);
+
+  auto prop_weight_map = boost::get(boost::edge_weight, graph);
+
+  std::map<UGraph::vertex_descriptor, UGraph::vertex_descriptor> predicessor_map;
+  boost::associative_property_map<std::map<UGraph::vertex_descriptor, UGraph::vertex_descriptor>> prop_predicessor_map(
+      predicessor_map);
+
+  std::map<UGraph::vertex_descriptor, double> distance_map;
+  boost::associative_property_map<std::map<UGraph::vertex_descriptor, double>> prop_distance_map(distance_map);
+
+  std::map<UGraph::vertex_descriptor, size_t> u_index_map;
+  boost::associative_property_map<std::map<UGraph::vertex_descriptor, size_t>> u_prop_index_map(u_index_map);
+
+  {  // Populate index map
+    int c = 0;
+    UGraph::vertex_iterator i, iend;
+    for (boost::tie(i, iend) = boost::vertices(graph); i != iend; ++i, ++c)
+    {
+      std::string name = boost::get(boost::vertex_link, graph)[*i]->getName();
+      if (name == root)
+        s_root = *i;
+
+      if (name == tip)
+        s_tip = *i;
+
+      boost::put(u_prop_index_map, *i, c);
+    }
+  }
 
   dijkstra_shortest_paths(graph,
-                          s,
+                          s_root,
                           prop_predicessor_map,
                           prop_distance_map,
                           prop_weight_map,
-                          prop_index_map,
+                          u_prop_index_map,
                           std::less<>(),
                           boost::closed_plus<double>(),
                           (std::numeric_limits<double>::max)(),
                           0,
                           boost::default_dijkstra_visitor());
 
-  std::vector<std::string> links;
-  std::vector<std::string> joints;
-  Vertex v = getVertex(tip);           // We want to start at the destination and work our way back to the source
-  for (Vertex u = predicessor_map[v];  // Start by setting 'u' to the destintaion node's predecessor
-       u != v;                         // Keep tracking the path until we get to the source
-       v = u, u = predicessor_map[v])  // Set the current vertex to the current predecessor, and the predecessor to one
-                                       // level up
+  ShortestPath path;
+  path.links.reserve(predicessor_map.size());
+  path.joints.reserve(predicessor_map.size());
+  path.active_joints.reserve(predicessor_map.size());
+  // We want to start at the destination and work our way back to the source
+  for (Vertex u = predicessor_map[s_tip];      // Start by setting 'u' to the destination node's predecessor
+       u != s_tip;                             // Keep tracking the path until we get to the source
+       s_tip = u, u = predicessor_map[s_tip])  // Set the current vertex to the current predecessor, and the predecessor
+                                               // to one level up
   {
-    links.push_back(boost::get(boost::vertex_link, graph)[v]->getName());
-    joints.push_back(boost::get(boost::edge_joint, graph)[boost::edge(u, v, graph).first]->getName());
+    path.links.push_back(boost::get(boost::vertex_link, graph)[s_tip]->getName());
+    const Joint::ConstPtr& joint = boost::get(boost::edge_joint, graph)[boost::edge(u, s_tip, graph).first];
+
+    path.joints.push_back(joint->getName());
+    if (joint->type != JointType::FIXED && joint->type != JointType::FLOATING)
+      path.active_joints.push_back(joint->getName());
   }
-  links.push_back(root);
-  std::reverse(links.begin(), links.end());
-  std::reverse(joints.begin(), joints.end());
+  path.links.push_back(root);
+  std::reverse(path.links.begin(), path.links.end());
+  std::reverse(path.joints.begin(), path.joints.end());
+  std::reverse(path.active_joints.begin(), path.active_joints.end());
 
 #ifndef NDEBUG
   CONSOLE_BRIDGE_logDebug("distances and parents:");
-  Graph::vertex_iterator vi, vend;
+  UGraph::vertex_iterator vi, vend;
   for (boost::tie(vi, vend) = boost::vertices(graph); vi != vend; ++vi)
   {
-    CONSOLE_BRIDGE_logDebug("distance(%s) = %d, parent(%s) = %s",
+    CONSOLE_BRIDGE_logDebug("distance(%s) = %f, parent(%s) = %s",
                             boost::get(boost::vertex_link, graph)[*vi]->getName().c_str(),
                             distance_map[*vi],
                             boost::get(boost::vertex_link, graph)[*vi]->getName().c_str(),
                             boost::get(boost::vertex_link, graph)[predicessor_map[*vi]]->getName().c_str());
   }
 #endif
-  return Path(links, joints);
+  return path;
 }
 
 SceneGraph::Vertex SceneGraph::getVertex(const std::string& name) const
 {
   auto found = link_map_.find(name);
   if (found == link_map_.end())
-    return Vertex();
+    throw std::runtime_error("SceneGraph, vertex with name '" + name + "' does not exist!");
 
   return found->second.second;
 }
@@ -662,7 +967,7 @@ SceneGraph::Edge SceneGraph::getEdge(const std::string& name) const
 {
   auto found = joint_map_.find(name);
   if (found == joint_map_.end())
-    return Edge{};
+    throw std::runtime_error("SceneGraph, edge with name '" + name + "' does not exist!");
 
   return found->second.second;
 }
@@ -670,12 +975,13 @@ SceneGraph::Edge SceneGraph::getEdge(const std::string& name) const
 /** addSceneGraph needs a couple helpers to handle prefixing, we hide them in an anonymous namespace here **/
 namespace
 {
-tesseract_scene_graph::Link clone_prefix(tesseract_scene_graph::Link::ConstPtr link, const std::string& prefix)
+tesseract_scene_graph::Link clone_prefix(const tesseract_scene_graph::Link::ConstPtr& link, const std::string& prefix)
 {
   return link->clone(prefix + link->getName());
 }
 
-tesseract_scene_graph::Joint clone_prefix(tesseract_scene_graph::Joint::ConstPtr joint, const std::string& prefix)
+tesseract_scene_graph::Joint clone_prefix(const tesseract_scene_graph::Joint::ConstPtr& joint,
+                                          const std::string& prefix)
 {
   auto ret = joint->clone(prefix + joint->getName());
   ret.child_link_name = prefix + joint->child_link_name;
@@ -683,12 +989,13 @@ tesseract_scene_graph::Joint clone_prefix(tesseract_scene_graph::Joint::ConstPtr
   return ret;
 }
 
-AllowedCollisionMatrix::Ptr clone_prefix(AllowedCollisionMatrix::ConstPtr acm, const std::string& prefix)
+tesseract_common::AllowedCollisionMatrix::Ptr
+clone_prefix(const tesseract_common::AllowedCollisionMatrix::ConstPtr& acm, const std::string& prefix)
 {
   if (prefix.empty())
-    return std::make_shared<AllowedCollisionMatrix>(*acm);
+    return std::make_shared<tesseract_common::AllowedCollisionMatrix>(*acm);
 
-  auto new_acm = std::make_shared<AllowedCollisionMatrix>();
+  auto new_acm = std::make_shared<tesseract_common::AllowedCollisionMatrix>();
   for (const auto& entry : acm->getAllAllowedCollisions())
     new_acm->addAllowedCollision(prefix + entry.first.first, prefix + entry.first.second, entry.second);
 
@@ -787,4 +1094,78 @@ bool SceneGraph::insertSceneGraph(const tesseract_scene_graph::SceneGraph& scene
   return addJointHelper(std::make_shared<Joint>(joint.clone()));
 }
 
+void SceneGraph::rebuildLinkAndJointMaps()
+{
+  link_map_.clear();
+  joint_map_.clear();
+
+  {  // Rebuild link map
+    Graph::vertex_iterator i, iend;
+    for (boost::tie(i, iend) = boost::vertices(*this); i != iend; ++i)
+    {
+      Link::Ptr link = boost::get(boost::vertex_link, *this)[*i];
+      link_map_[link->getName()] = std::make_pair(link, *i);
+    }
+  }
+
+  {  // Rebuild joint map
+    Graph::edge_iterator i, iend;
+    for (boost::tie(i, iend) = boost::edges(*this); i != iend; ++i)
+    {
+      Joint::Ptr joint = boost::get(boost::edge_joint, *this)[*i];
+      joint_map_[joint->getName()] = std::make_pair(joint, *i);
+    }
+  }
+}
+
+bool SceneGraph::operator==(const SceneGraph& rhs) const
+{
+  using namespace tesseract_common;
+  // Currently these only compare the Link/Joint.
+  auto link_pair_equal = [](const std::pair<const Link::Ptr, Vertex>& v1, const std::pair<Link::Ptr, Vertex>& v2) {
+    return pointersEqual(v1.first, v2.first);
+  };
+  auto joint_pair_equal = [](const std::pair<const Joint::Ptr, Edge>& v1, const std::pair<Joint::Ptr, Edge>& v2) {
+    return pointersEqual(v1.first, v2.first);
+  };
+
+  bool equal = true;
+  equal &= pointersEqual(acm_, rhs.acm_);
+  equal &= isIdenticalMap<std::unordered_map<std::string, std::pair<Link::Ptr, Vertex>>, std::pair<Link::Ptr, Vertex>>(
+      link_map_, rhs.link_map_, link_pair_equal);
+  equal &= isIdenticalMap<std::unordered_map<std::string, std::pair<Joint::Ptr, Edge>>, std::pair<Joint::Ptr, Edge>>(
+      joint_map_, rhs.joint_map_, joint_pair_equal);
+
+  return equal;
+}
+bool SceneGraph::operator!=(const SceneGraph& rhs) const { return !operator==(rhs); }
+
+template <class Archive>
+void SceneGraph::save(Archive& ar, const unsigned int /*version*/) const
+{
+  using namespace boost::serialization;
+  ar& BOOST_SERIALIZATION_BASE_OBJECT_NVP(Graph);
+  ar& BOOST_SERIALIZATION_NVP(acm_);
+}
+
+template <class Archive>
+void SceneGraph::load(Archive& ar, const unsigned int /*version*/)
+{
+  using namespace boost::serialization;
+  ar& BOOST_SERIALIZATION_BASE_OBJECT_NVP(Graph);
+  ar& BOOST_SERIALIZATION_NVP(acm_);
+
+  rebuildLinkAndJointMaps();
+}
+
+template <class Archive>
+void SceneGraph::serialize(Archive& ar, const unsigned int version)
+{
+  boost::serialization::split_member(ar, *this, version);
+}
+
 }  // namespace tesseract_scene_graph
+
+#include <tesseract_common/serialization.h>
+TESSERACT_SERIALIZE_SAVE_LOAD_ARCHIVES_INSTANTIATE(tesseract_scene_graph::SceneGraph)
+BOOST_CLASS_EXPORT_IMPLEMENT(tesseract_scene_graph::SceneGraph)
